#include "memmgr.h"

char *stateStr[3]={"FREE","OCCUPIED"};

SequentialFitMemoryManager::SequentialFitMemoryManager(U4 nbytes)
{
	handle = GetProcessHeap();
	if(handle==NULL)
	{
		printf("SequentialFitMemoryManager::");
		printf("SequentialFitMemoryManager():");
		printf("invalid handle\n");
		exit(1);
	}

	ram = (U1*)HeapAlloc(handle,HEAP_ZERO_MEMORY,nbytes);
	
	//for portability, you could use:
	//ram = (unsigned char*)malloc(nbytes);


	size = nbytes;

	if(size<=SZ_HEADER)
	{
		printf("SequentialFitMemoryManager::");
		printf("SequentialFitMemoryManager():");
		printf("not enough memory fed to constructor\n");
		exit(1);	
	}

	PREV(START)=0;
	NEXT(START)=0;
	STATE(START)=FREE;
	SIZE(START)=size-SZ_HEADER;

	MSG0("SequentialFitMemoryManager::");
	MSG1("SequentialFitMemoryManager(%lu)\n",nbytes);

	return;

}/*end constructor-------------------------------------------------*/

SequentialFitMemoryManager::~SequentialFitMemoryManager()
{
	if(HeapFree(handle,HEAP_NO_SERIALIZE,ram)==0)
	{
		printf("SequentialFitMemoryManager::");
		printf("~SequentialFitMemoryManager():");
		printf("could not free heap storage\n");
		return;
	}

	//for portability, you could use:
	//free(ram);

	MSG0("SequentialFitMemoryManager::");
	MSG0("~SequentialFitMemoryManager()");
	MSG1("free ram[%lu]\n",size);
	return;

}/*end destructor--------------------------------------------------*/

/*
U4 nbytes	-	number of bytes required
returns address of first byte of memory region allocated 
( or NULL if cannot allocate a large enough block )
*/

void* SequentialFitMemoryManager::allocate(U4 nbytes)
{
	U4 current;

	MSG0("SequentialFitMemoryManager::");
	MSG1("allocate(%lu)\n",nbytes);

	if(nbytes==0)
	{ 
		MSG0("SequentialFitMemoryManager::");
		MSG0("allocate(): zero bytes requested\n");
		return(NULL);
	}

	//traverse the linked list, starting with first element

	current = START;
	while(NEXT(current)!=0)
	{
		if((SIZE(current)>=nbytes)&&(STATE(current)==FREE))
		{
			split(current,nbytes);
			return((void*)&ram[current]);
		}
		current = NEXT(current);
	}

	//handle the last block ( which has NEXT(current)=0 )

	if((SIZE(current)>=nbytes)&&(STATE(current)==FREE))
	{
		split(current,nbytes);
		return((void*)&ram[current]);
	}

	return(NULL);

}/*end allocation--------------------------------------------------*/

/*
breaks [free] region into [alloc][free] pair, if possible
*/

void SequentialFitMemoryManager::split(U4 addr, U4 nbytes)
{
	/*
	want payload to have enough room for
	nbytes = size of request
	SZ_HEADER = header for new region
	SZ_HEADER = payload for new region (arbitrary 13 bytes)
	*/

	if(SIZE(addr)>= nbytes+SZ_HEADER+SZ_HEADER)
	{
		U4 oldnext;
		U4 oldprev;
		U4 oldsize;

		U4 newaddr;

		MSG0("SequentialFitMemoryManager::");
		MSG0("split(): split=YES\n");

		oldnext=NEXT(addr);
		oldprev=PREV(addr);
		oldsize=SIZE(addr);

		newaddr = addr + nbytes + SZ_HEADER;
		
		NEXT(addr)=newaddr;
		PREV(addr)=oldprev;
		STATE(addr)=OCCUPIED;
		SIZE(addr)=nbytes;

		NEXT(newaddr)=oldnext;
		PREV(newaddr)=addr;
		STATE(newaddr)=FREE;
		SIZE(newaddr)=oldsize-nbytes-SZ_HEADER;
	}
	else 
	{
		MSG0("SequentialFitMemoryManager::");
		MSG0("split(): split=NO\n");
		STATE(addr)=OCCUPIED;
	}
	return;

}/*end split-------------------------------------------------------*/

void SequentialFitMemoryManager::release(void *addr)
{
	U4 free;	//index into ram[]

	if(addr==NULL)
	{
		MSG0("SequentialFitMemoryManager::");
		MSG0("release(): cannot release NULL pointer\n");
		return;
	}

	MSG0("SequentialFitMemoryManager::");
	MSG1("release(%lu)\n",addr);

	//perform sanity check to make sure address is kosher

	if( (addr>= (void*)&ram[size]) || (addr< (void*)&ram[0]) )
	{
		MSG0("SequentialFitMemoryManager::");
		MSG0("release(): address out of bounds\n");
		return;
	}

	//translate void* addr to index in ram[]

	free = (U4)( ((U1*)addr) - &ram[0] );

	MSG0("SequentialFitMemoryManager::");
	MSG1("address resolves to index %lu\n",free);

	//a header always occupies first 13 bytes of storage

	if(free<13)
	{
		MSG0("SequentialFitMemoryManager::");
		MSG0("release(): address in first 13 bytes\n");
		return;	
	}

	//yet more sanity checks

	if((STATE(free)!=OCCUPIED)	||	//region if free
	   (PREV(free)>=free)		||	//previous element not previous
	   (NEXT(free)>=size)		||	//next is beyond the end
	   (SIZE(free)>=size)		||	//size of region greater than whole
	   (SIZE(free)==0))				//no size at all
	{
		MSG0("SequentialFitMemoryManager::");
		MSG0("release(): referencing invalid region\n");
		return;	
	}

	merge(PREV(free),free,NEXT(free));

	return;

}/*end release-----------------------------------------------------*/

/*
		4 cases ( F=free O=occupied )
			FOF	-> [F]
			OOF -> O[F]
			FOO -> [F]O
			OOO -> OFO
*/

void SequentialFitMemoryManager::merge(U4 prev,U4 current,U4 next)
{
	/*
	first handle special cases of region at end(s)
	prev=0					low end
	next=0					high end
	prev=0 and next=0		only 1 list element
	*/

	if(prev==0)
	{
		if(next==0)
		{
			STATE(current)=FREE;
		}
		else if(STATE(next)==OCCUPIED)
		{
			STATE(current)=FREE;
		}
		else if(STATE(next)==FREE)
		{
			U4 temp;

			MSG0("SequentialFitMemoryManager::merge():");
			MSG0("merging to NEXT\n");

			STATE(current)=FREE;
			SIZE(current)=SIZE(current)+SIZE(next)+SZ_HEADER;
			NEXT(current)=NEXT(next);

			temp = NEXT(next);
			if(temp!=0){ PREV(temp)=current; }
		}
	}
	else if(next==0)
	{
		if(STATE(prev)==OCCUPIED)
		{
			STATE(current)=FREE;
		}
		else if(STATE(prev)==FREE)
		{
			MSG0("SequentialFitMemoryManager::merge():");
			MSG0("merging to PREV\n");
			SIZE(prev)=SIZE(prev)+SIZE(current)+SZ_HEADER;
			NEXT(prev)=NEXT(current);
		}
	}
	
	/* now we handle 4 cases */

	else if((STATE(prev)==OCCUPIED)&&(STATE(next)==OCCUPIED))
	{
		STATE(current)=FREE;
	}
	else if((STATE(prev)==OCCUPIED)&&(STATE(next)==FREE))
	{
		U4 temp;

		MSG0("SequentialFitMemoryManager::merge():");
		MSG0("merging to NEXT\n");
		STATE(current)=FREE;
		SIZE(current)=SIZE(current)+SIZE(next)+SZ_HEADER;
		NEXT(current)=NEXT(next);

		temp = NEXT(next);
		if(temp!=0){ PREV(temp)=current; }
	}
	else if((STATE(prev)==FREE)&&(STATE(next)==OCCUPIED))
	{
		MSG0("SequentialFitMemoryManager::merge():");
		MSG0("merging to PREV\n");
		SIZE(prev)=SIZE(prev)+SIZE(current)+SZ_HEADER;
		NEXT(prev)=NEXT(current);
		PREV(next)=prev;
	}
	else if((STATE(prev)==FREE)&&(STATE(next)==FREE))
	{
		U4 temp;

		MSG0("SequentialFitMemoryManager::merge():");
		MSG0("merging with both sides\n");
		SIZE(prev)=SIZE(prev)+
			       SIZE(current)+SZ_HEADER+
				   SIZE(next)+SZ_HEADER;
		NEXT(prev)=NEXT(next);

		temp = NEXT(next);
		if(temp!=0){ PREV(temp)=prev; }
	}

	return;

}/*end merge-------------------------------------------------------*/

void SequentialFitMemoryManager::printState()
{
	U4 i;
	U4 current;

	i=0;
	current=START;
	
	while(NEXT(current)!=0)
	{
		printf("%lu) [P=%lu]",i,PREV(current));
		printf("[addr=%lu]",current);
		printf("[St=%s]",stateStr[STATE(current)]);
		printf("[Sz=%lu]",SIZE(current));
		printf("[N=%lu]\n",NEXT(current));
		current = NEXT(current);
		i++;
	}

	//print the last list element

	printf("%lu) [P=%lu]",i,PREV(current));
	printf("[addr=%lu]",current);
	printf("[St=%s]",stateStr[STATE(current)]);
	printf("[Sz=%lu]",SIZE(current));
	printf("[N=%lu]\n",NEXT(current));

	return;

}/*end printState--------------------------------------------------*/

